import GLib from "gi://GLib";
import Gio from "gi://Gio";
import { Extension, InjectionManager, gettext as _ } from "resource:///org/gnome/shell/extensions/extension.js";
import * as Main from "resource:///org/gnome/shell/ui/main.js";
import * as PopupMenu from "resource:///org/gnome/shell/ui/popupMenu.js";
import WallpaperQueue from "./helpers/shell/WallpaperQueue.js";
import { parseDynamicWallpaper } from "./utils/common/dwp.js";
import { readFile } from "./utils/common/io.js";
import { debugLog, errorLog, handleCatch, isBitSet } from "./utils/common/misc.js";

Gio._promisify(Gio.File.prototype, "enumerate_children_async", "enumerate_children_finish");
Gio._promisify(Gio.File.prototype, "query_info_async", "query_info_finish");
Gio._promisify(Gio.FileEnumerator.prototype, "next_files_async", "next_files_finish");

export default class Wallhub extends Extension {
    settings;
    backgroundSettings;
    wallpaperPaths;
    wallpaperPathsSelected;
    slideshowIntervalUnit;
    slideshowInterval;
    wallpaperQueue;
    directoryMonitors;
    loopSourceId;
    injectionManager;
    enable() {
        this.settings = this.getSettings();
        this.backgroundSettings = new Gio.Settings({ schema: "org.gnome.desktop.background" });
        this.wallpaperPaths = this.settings.get_strv("wallpaper-paths");
        this.wallpaperPathsSelected = this.settings.get_int("wallpaper-paths-selected");
        this.slideshowIntervalUnit = this.settings.get_enum("slideshow-interval-unit");
        this.slideshowInterval = this.settings.get_uint("slideshow-interval");
        this.wallpaperQueue = new WallpaperQueue();
        this.directoryMonitors = [];
        this.injectionManager = new InjectionManager();

        this.settings.connect("changed::wallpaper-paths", () => {
            this.wallpaperPaths = this.settings.get_strv("wallpaper-paths");
        });

        this.settings.connect("changed::wallpaper-paths-selected", () => {
            this.wallpaperPathsSelected = this.settings.get_int("wallpaper-paths-selected");
            this.updateQueue().catch(handleCatch);
        });

        this.settings.connect("changed::slideshow-interval-unit", () => {
            this.slideshowIntervalUnit = this.settings.get_enum("slideshow-interval-unit");
            this.startLoop().catch(handleCatch);
        });

        this.settings.connect("changed::slideshow-interval", () => {
            this.slideshowInterval = this.settings.get_uint("slideshow-interval");
            this.startLoop().catch(handleCatch);
        });

        this.injectionManager.overrideMethod(Main.layoutManager, "_addBackgroundMenu", (originalMethod) => {
            // eslint-disable-next-line @typescript-eslint/no-this-alias
            return (bgManager) => {
                debugLog("Overriding _addBackgroundMenu");
                originalMethod.call(Main.layoutManager, bgManager);
                // @ts-expect-error _backgroundMenu is private
                const menu = bgManager.backgroundActor._backgroundMenu;
                this.addBackgroundMenuItem(menu);
            };
        });

        this.updateQueue() //
            .then(this.startLoop.bind(this))
            .catch(handleCatch);

        debugLog("Enabled");
    }
    disable() {
        this.settings = null;
        this.backgroundSettings = null;
        this.wallpaperPaths = null;
        this.wallpaperPathsSelected = null;
        this.slideshowIntervalUnit = null;
        this.slideshowInterval = null;
        this.wallpaperQueue.clear();
        this.directoryMonitors.forEach((monitor) => monitor.cancel());
        this.injectionManager.clear();
        this.wallpaperQueue = null;
        this.directoryMonitors = null;
        this.injectionManager = null;

        // @ts-expect-error _bgManagers is private
        Main.layoutManager._bgManagers.forEach((bgManager) => {
            // @ts-expect-error _backgroundMenu is private
            const menu = bgManager.backgroundActor._backgroundMenu;

            // @ts-expect-error _getMenuItems is private
            menu._getMenuItems().forEach((item) => {
                // @ts-expect-error label is not in types
                if (item.label_actor?.text === _("Next Wallpaper")) {
                    menu.box.remove_child(item.actor);
                }
            });
        });

        if (this.loopSourceId != null) {
            GLib.source_remove(this.loopSourceId);
        }

        debugLog("Disabled");
    }
    async startLoop() {
        if (this.loopSourceId > 0) {
            GLib.source_remove(this.loopSourceId);
        }

        if (this.wallpaperQueue.length === 0) {
            return;
        }
        else {
            const interval = this.slideshowInterval * this.slideshowIntervalUnit;

            this.loopSourceId = GLib.timeout_add_seconds(GLib.PRIORITY_DEFAULT, interval, () => {
                const wallpaper = this.wallpaperQueue.dequeue();

                if (wallpaper != null) {
                    this.setWallpaper(wallpaper).catch(handleCatch);
                    return GLib.SOURCE_CONTINUE;
                }

                this.loopSourceId = null;
                return GLib.SOURCE_REMOVE;
            });
        }
    }
    async updateQueue() {
        this.wallpaperQueue.clear();

        for (let i = 0; i < this.wallpaperPaths.length; i++) {
            const path = this.wallpaperPaths[i];

            if (isBitSet(this.wallpaperPathsSelected, i) === false) {
                continue;
            }

            const file = Gio.File.new_for_path(path);
            const fileType = file.query_file_type(Gio.FileQueryInfoFlags.NONE, null);

            if (fileType === Gio.FileType.REGULAR) {
                this.wallpaperQueue.enqueue(path);
            }
            else if (fileType === Gio.FileType.DIRECTORY) {
                const wallpapers = await this.getWallpapersFromDirectory(file).catch(handleCatch);

                if (wallpapers == null) {
                    errorLog("Failed to get wallpapers from directory", path);
                    return;
                }

                for (const wallpaper of wallpapers) {
                    this.wallpaperQueue.enqueue(wallpaper);
                }
            }
            else {
                errorLog(`Invalid file type: ${path}`);
            }
        }
    }
    async getWallpapersFromDirectory(directory) {
        const enumerator = await directory.enumerate_children_async("standard::*", Gio.FileQueryInfoFlags.NONE, GLib.PRIORITY_DEFAULT, null);
        const wallpapers = [];

        while (true) {
            const fileInfos = await enumerator.next_files_async(10, GLib.PRIORITY_DEFAULT, null);
            if (fileInfos.length === 0)
                break;

            for (const fileInfo of fileInfos) {
                const file = directory.get_child(fileInfo.get_name());
                const fileType = fileInfo.get_file_type();

                if (fileType === Gio.FileType.REGULAR) {
                    wallpapers.push(file.get_path());
                }
                else if (fileType === Gio.FileType.DIRECTORY) {
                    const childWallpapers = await this.getWallpapersFromDirectory(file);
                    wallpapers.push(...childWallpapers);
                }
            }
        }

        await this.createDirectoryMonitor(directory).catch(handleCatch);
        return wallpapers;
    }
    async createDirectoryMonitor(file) {
        const monitor = file.monitor_directory(Gio.FileMonitorFlags.NONE, null);
        monitor.rateLimit = 1000;

        monitor.connect("changed", (monitor, file, otherFile, eventType) => {
            if (eventType === Gio.FileMonitorEvent.CREATED) {
                const pathname = file.get_path();

                if (this.wallpaperQueue.includes(pathname) === false) {
                    this.wallpaperQueue.enqueue(pathname);
                }
            }
            else if (eventType === Gio.FileMonitorEvent.DELETED) {
                this.wallpaperQueue.remove(file.get_path());
            }
        });

        this.directoryMonitors.push(monitor);
    }
    async setWallpaper(path) {
        const extension = path.split(".").pop().toLowerCase();

        if (extension === "xml") {
            const xmlBytes = await readFile(path, null);

            if (xmlBytes == null) {
                errorLog("Failed to read dynamic wallpaper");
                return;
            }

            const decoder = new TextDecoder();
            const xml = decoder.decode(xmlBytes);
            const dwpConfig = parseDynamicWallpaper(xml);

            if (dwpConfig == null) {
                errorLog("Failed to parse dynamic wallpaper");
                return;
            }

            this.backgroundSettings.set_string("picture-uri", `file:///${dwpConfig.lightBg}`);
            this.backgroundSettings.set_string("picture-uri-dark", `file:///${dwpConfig.darkBg}`);
        }
        else {
            this.backgroundSettings.set_string("picture-uri", `file:///${path}`);
            this.backgroundSettings.set_string("picture-uri-dark", `file:///${path}`);
        }
    }
    addBackgroundMenuItem(menu) {
        const menuItem = new PopupMenu.PopupMenuItem(_("Next Wallpaper"));

        menuItem.connect("activate", () => {
            const wallpaper = this.wallpaperQueue.dequeue();

            if (wallpaper != null) {
                this.setWallpaper(wallpaper).catch(handleCatch);
            }
        });

        menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem(null));
        menu.addMenuItem(menuItem);
    }
}
